function demo_Gaussian03
% Gaussian conditioning with uncertain inputs
%
% If this code is useful for your research, please cite the related publication:
% @incollection{Calinon19MM,
% 	author="Calinon, S.",
% 	title="Mixture Models for the Analysis, Edition, and Synthesis of Continuous Time Series",
% 	booktitle="Mixture Models and Applications",
% 	publisher="Springer",
% 	editor="Bouguila, N. and Fan, W.", 
% 	year="2019",
% 	pages="39--57",
% 	doi="10.1007/978-3-030-23876-6_3"
% }
%
% Copyright (c) 2019 Idiap Research Institute, http://idiap.ch/
% Written by Sylvain Calinon, http://calinon.ch/
% 
% This file is part of PbDlib, http://www.idiap.ch/software/pbdlib/
% 
% PbDlib is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License version 3 as
% published by the Free Software Foundation.
% 
% PbDlib is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with PbDlib. If not, see <http://www.gnu.org/licenses/>.

addpath('./m_fcts/');


%% Parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
model.nbVar = 2; %Number of variables [x1,x2]
nbSamples = 10000; %Number of stochastic samples


%% Load  data
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
load('data/faithful.mat');
Data = faithful';
Data(1,:) = Data(1,:)*1E1;


%% Gaussian conditioning with uncertain inputs
%% (For Gaussian conditioning, see for example section "2.3.1 Conditional Gaussian distributions" in Bishop's book, 
%% or the "Conditional distribution" section on the multivariate normal distribution wikipedia page)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
model.Mu = mean(Data,2);
model.Sigma = cov(Data');

in=1; out=2;
DataIn = 50;
SigmaIn = 1E1;

MuOut = model.Mu(out) + model.Sigma(out,in) / (model.Sigma(in,in) + SigmaIn) * (DataIn - model.Mu(in));
SigmaOut = model.Sigma(out,out) - model.Sigma(out,in) / (model.Sigma(in,in) + SigmaIn) * model.Sigma(in,out);
slope = model.Sigma(out,in) / (model.Sigma(in,in) + SigmaIn);


%% Gaussian conditioning generated by a Gaussian product followed by stochastically sampled data on input and output
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Compute product of Gaussians between N(xIn,SigmaIn) and N(Mu(in),Sigma(in,in)) and store it in N(xP,SigmaP)
SigmaTmp = inv(SigmaIn) + inv(model.Sigma(in,in));
MuTmp = SigmaIn \ DataIn + model.Sigma(in,in) \ model.Mu(in);
SigmaP = inv(SigmaTmp);
xP = SigmaP * MuTmp;

% %Stochastic sampling on the input
% [V,D] = eig(SigmaP);
% xIn = repmat(xP,1,nbSamples) + V * D^.5 * randn(1,nbSamples);

% %Regression
% xMuOut = repmat(model.Mu(out),1,nbSamples) + model.Sigma(out,in) / model.Sigma(in,in) * (xIn - repmat(model.Mu(in),1,nbSamples));
xSigmaOut = model.Sigma(out,out) - model.Sigma(out,in) / model.Sigma(in,in) * model.Sigma(in,out);


% %Stochastic sampling on the output
% [V,D] = eig(xSigmaOut);
% xOut = xMuOut + V * D^.5 * randn(1,nbSamples);
% 
% %Gaussian fit on the stochastic samples
% MuOut2 = mean(xOut,2);
% SigmaOut2 = cov(xOut');

%Regression
A = model.Sigma(out,in) / model.Sigma(in,in);
MuOut2 = repmat(model.Mu(out),1,nbSamples) + A * (xP - repmat(model.Mu(in),1,nbSamples));
SigmaOut2 = xSigmaOut + A * SigmaP * A';


% %% Gaussian conditioning with uncertain inputs, via sampling
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %Stochastic sampling on the input
% [V,D] = eig(SigmaIn);
% xIn = repmat(DataIn,1,nbSamples) + V * D^.5 * randn(1,nbSamples);
% 
% %Regression
% xMuOut = repmat(model.Mu(out),1,nbSamples) + model.Sigma(out,in) / model.Sigma(in,in) * (xIn - repmat(model.Mu(in),1,nbSamples));
% xSigmaOut = model.Sigma(out,out) - model.Sigma(out,in) / model.Sigma(in,in) * model.Sigma(in,out);
% 
% %Stochastic sampling on the output
% [V,D] = eig(xSigmaOut);
% xOut = xMuOut + V * D^.5 * randn(1,nbSamples);
% 
% %Gaussian fit on the stochastic samples
% MuOut3 = mean(xOut,2);
% SigmaOut3 = cov(xOut');

A = model.Sigma(out,in) / model.Sigma(in,in);
MuOut3 = repmat(model.Mu(out),1,nbSamples) + A * (DataIn - repmat(model.Mu(in),1,nbSamples));
SigmaOut3 = xSigmaOut + A * SigmaIn * A';


%% Plot
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
mrg = [10 10];
limAxes = [min(Data(1,:))-mrg(1) max(Data(1,:))+mrg(1) min(Data(2,:))-mrg(2) max(Data(2,:))+mrg(2)];

figure('PaperPosition',[0 0 4 3],'position',[10,10,1200,800]); hold on; %axis off;
plot(DataIn+[-50,50], MuOut+slope*[-50,50], ':','linewidth',1,'color',[.7 .3 .3]);
plot([model.Mu(1) model.Mu(1)], [limAxes(3) model.Mu(2)], ':','linewidth',1,'color',[.7 .3 .3]);
plot([limAxes(1) model.Mu(1)], [model.Mu(2) model.Mu(2)], ':','linewidth',1,'color',[.7 .3 .3]);

%Plot joint distribution
plotGMM(model.Mu, model.Sigma, [.8 0 0], .4);
%Plot marginal distribution horizontally
plotGaussian1D(model.Mu(1), model.Sigma(1,1), [limAxes(1) limAxes(3) limAxes(2)-limAxes(1) 10], [.8 0 0], .4, 'h');
%Plot marginal distribution vertically
plotGaussian1D(model.Mu(2), model.Sigma(2,2), [limAxes(1) limAxes(3) 10 limAxes(4)-limAxes(3)], [.8 0 0], .4, 'v');
%Plot input distribution horizontally
plotGaussian1D(DataIn, SigmaIn, [limAxes(1) limAxes(3) limAxes(2)-limAxes(1) 10], [0 0 .8], .4, 'h');
%Plot conditional distribution vertically
[~,h(1,:)] = plotGaussian1D(MuOut, SigmaOut, [limAxes(1) limAxes(3) 10 limAxes(4)-limAxes(3)], [0 0 .8], .4, 'v');
%Plot estimated output distribution from stochastic sampling
[~,h(2,:)] = plotGaussian1D(MuOut2, SigmaOut2, [limAxes(1) limAxes(3) 10 limAxes(4)-limAxes(3)], [0 .8 0], .4, 'v');
%Plot Gaussian conditioning with uncertain inputs, via sampling
[~,h(3,:)] = plotGaussian1D(MuOut3, SigmaOut3, [limAxes(1) limAxes(3) 10 limAxes(4)-limAxes(3)], [.5 .5 .5], .4, 'v');

plot(DataIn,MuOut,'.','markersize',12,'color',[.7 .3 .3]);
plot(DataIn,limAxes(3),'.','markersize',12,'color',[.7 .3 .3]);
plot([DataIn DataIn], [limAxes(3) MuOut], ':','linewidth',1,'color',[.7 .3 .3]);
plot([limAxes(1) DataIn], [MuOut MuOut], ':','linewidth',1,'color',[.7 .3 .3]);

axis(limAxes);
set(gca,'Xtick',[]); set(gca,'Ytick',[]);
%axis equal;
%plot(Data(1,1:200), Data(2,1:200), 'o','markersize',4,'color',[.5 .5 .5]);
xlabel('$x_1$','fontsize',16,'interpreter','latex');
ylabel('$x_2$','fontsize',16,'interpreter','latex');
legend(h(:,1), {'Proposed formulation','Product with marginal input distributions then sampling','sampling'});

%print('-dpng','-r600','graphs/demo_Gaussian03.png');
pause;
close all;